import smtplib
from email.message import EmailMessage

CIRTAP_MAILER = "cirtap@mydomain.com"


def send_start_mail(
    recipients,
    db_dir,
    no_of_genomes,
):
    msg = EmailMessage()

    msg["Subject"] = "[CIRTAP-BOT] PATRIC update"
    msg["From"] = CIRTAP_MAILER
    msg["To"] = recipients

    start_content = (
        "A mirror job for PATRIC was launched by cirtap.\n"
        "Details:\n"
        "\n"
        "Local dir: {}\n"
        "No. of genomes: {}\n"
        "\n"
        "You may receive another email once the run finishes successfully or "
        "fails\n"
    ).format(db_dir.resolve(), no_of_genomes)

    msg.set_content(start_content)

    s = smtplib.SMTP("localhost")
    s.sendmail(CIRTAP_MAILER, recipients, msg.as_string())
    s.quit()


def send_exit_mail(recipients, error_msg=None):
    msg = EmailMessage()

    msg["Subject"] = "[CIRTAP-BOT] PATRIC update"
    msg["From"] = CIRTAP_MAILER
    msg["To"] = recipients

    if error_msg:
        msg_pre = "Cirtap failed!"
        error_content = "Error was : {}".format(error_msg)

    else:
        msg_pre = "Cirtap succeeded!"
        error_content = ""

    message_txt = "\n".join([msg_pre, error_content])
    msg.set_content(message_txt)
    s = smtplib.SMTP("localhost")
    s.sendmail(CIRTAP_MAILER, recipients, msg.as_string())
    s.quit()



import logging
import pathlib
import numpy as np
import pandas as pd

from ete3 import NCBITaxa

# CONSTANTS
OFFICIAL_RANKS = [
    "superkingdom",
    "phylum",
    "class",
    "order",
    "family",
    "genus",
    "species",
]

column_dtypes = {
    "taxid": "string",
    "taxon_id": "string",
    "genome_id": "string",
}

_logger = logging.getLogger(__name__)


def field_to_float(field_string):
    """Convert a str representation of a float to a float"""
    if field_string != "":
        return float(field_string)
    else:
        return np.nan


def parse_genome_summary(genome_summary_fp, thresh, available_genomes):
    """Parse the genome summary file in a data frame

    Only genomes passing the completeness - 5*contamination > thresh
    rule are retained.

    Positional arguments:
      genome_summary_fp: pathlib.Path: Path to the genome_summary file
      thresh: int: Threshold used for filtering out based on the rule
      available_genomes: list: List of strings of genome ids that have
        a genome file

    Return
      df: pd.DataFrame: A data frame with the stats per genomes that
        pass this first filtering
    """
    data = []
    with open(genome_summary_fp, "r") as fin:
        next(fin)
        for line in fin:
            fields = line.split("\t")
            genome_id = fields[0]
            genome_name = fields[1]
            taxid = fields[2]
            genome_status = fields[3].lower()
            coarse_consistency = field_to_float(fields[14])
            fine_consistency = field_to_float(fields[15])
            completeness = field_to_float(fields[16])
            contamination = field_to_float(fields[17])

            if (
                (completeness and contamination)
                and (completeness - (5 * contamination) > thresh)
                and (genome_status in ["complete", "wgs"])
            ):
                genome_stats = {
                    "genome_id": genome_id,
                    "genome_name": genome_name,
                    "taxid": taxid,
                    "genome_status": genome_status,
                    "completeness": completeness,
                    "contamination": contamination,
                    "coarse_consistency": coarse_consistency,
                    "fine_consistency": fine_consistency,
                }

                data.append(genome_stats)
            else:
                pass

    df = pd.DataFrame.from_records(data)

    _logger.info("Passing genomes : {}".format(genome_summary_fp, df.shape[0]))
    # Convert taxid and genome_id types to string
    # Plays better with the NCBI tree
    df[["taxid", "genome_id"]] = df[["taxid", "genome_id"]].astype(str)

    valid_genomes = set(available_genomes)
    df = df.loc[df.genome_id.isin(valid_genomes)]

    if df.shape[0] == 0:
        raise ValueError("Ooops. Something went wrong with filtering")

    return df


### FILTERING RULES ###
def filter_df_on_completeness(in_df):
    """Get a df or string of the genome id with the highest completeness"""
    max_completeness = in_df.completeness.max()
    on_completeness = in_df.loc[
        in_df.completeness == max_completeness,
    ]
    if on_completeness.shape[0] == 1:
        return on_completeness.genome_id.values[0]
    else:
        return on_completeness


def filter_df_on_contamination(in_df):
    """Get a df or string of the genome id with the lowest contamination"""
    min_contamination = in_df.contamination.min()
    on_contamination = in_df.loc[
        in_df.contamination == min_contamination,
    ]
    if on_contamination.shape[0] == 1:
        return on_contamination.genome_id.values[0]
    else:
        return on_contamination


# TO DO
# Make a function for coarse consistency
# Once the genome_summary gets fixed
def filter_df_on_fine_consistency(in_df):
    """Get a df or string of the genome id with the highest fine consistency"""

    max_fine_consistency = in_df.fine_consistency.min()
    on_fine_consistency = in_df.loc[
        in_df.fine_consistency == max_fine_consistency,
    ]
    if on_fine_consistency.shape[0] == 1:
        return on_fine_consistency.genome_id.values[0]
    else:
        return on_fine_consistency


def filter_df_on_status(in_df):
    """Get a df or string of the genome id with status complete or wgs"""
    statuses = in_df.genome_status.unique()

    if "complete" in statuses:
        on_status = in_df.loc[in_df.genome_status == "complete"]
    else:
        on_status = in_df.loc[in_df.genome_status == "wgs"]

    if on_status.shape[0] == 1:
        return on_status.genome_id.values[0]
    else:
        return on_status


def final_selection(status_df, random_state=1234):
    """Get the genome id as string no matter what.

    If we reach this step there must be a string instance returned.
    This is done with sampling one genome_id at random.
    The random_state ensures that we always choose the same genome from the
    same dataframe.

    Positional arguments:
      status_df: pd.DataFrame: A dataframe that has at least genome_id as
      column

    Optional arguments:
      random_state: int: An integer that controls the random state of
        pd.DataFrame.sample()

    Return:
      genome_id : string: A string of the `genome_id`
    """
    status_df = status_df.sort_values(by="genome_id")
    genome_id = status_df.sample(
        1, random_state=random_state
    ).genome_id.values[0]
    return genome_id


def get_best_genome_id(df):
    """Filter the df until a genome id is returned as string"""
    genome_id = np.nan

    while True:
        genome_id = filter_df_on_completeness(df)
        if type(genome_id) == str:
            break

        genome_id = filter_df_on_contamination(genome_id)
        if type(genome_id) == str:
            break

        genome_id = filter_df_on_fine_consistency(genome_id)
        if type(genome_id) == str:
            break

        genome_id = filter_df_on_status(genome_id)
        if type(genome_id) == str:
            break

        genome_id = final_selection(genome_id)
        if type(genome_id) == str:
            break

        if type(genome_id) != str:
            _logger.error(
                "No valid genome_id found for taxid: {}".format(genome_id)
            )
            _logger.error("Last data seen for it: \n{}".format(genome_id))
            break

    return genome_id


def select_best_from_tree(tree, best_per_taxid_df):
    """Select the best genome from a TreeNode instance"""
    genome_id = np.nan

    # Edge case for species-level tree
    if tree.is_leaf():
        taxids = [tree.name]
    else:
        taxids = [d.name for d in tree.iter_descendants()]

    subset_df = best_per_taxid_df.loc[best_per_taxid_df.taxid.isin(taxids)]

    if not subset_df.empty:
        genome_id = get_best_genome_id(subset_df)
    else:
        _logger.debug(
            "No genome available for tree:{} (rank: {}, taxid: {})".format(
                tree.sci_name, tree.rank, tree.name
            )
        )

    return genome_id


def select_best_for_rank(rank_name, ncbi_tree, best_per_taxid_df, lineages_df):
    """Create a dataframe that holds best genomes for a specified rank"""

    rank_trees = ncbi_tree.search_nodes(rank=rank_name)

    genome_ids = []
    for tree in rank_trees:
        genome_id = select_best_from_tree(tree, best_per_taxid_df)
        if genome_id:
            genome_ids.append(genome_id)

    rank_df = best_per_taxid_df.loc[
        best_per_taxid_df.genome_id.isin(genome_ids)
    ]

    rank_lineages = lineages_df.loc[lineages_df.genome_id.isin(genome_ids)]

    rank_complete_df = pd.merge(rank_df, rank_lineages, on="genome_id")

    _logger.info("{} : {}".format(rank_name, rank_complete_df.shape[0]))

    return rank_complete_df


def write_rank_df_to_file(rank_df, output_tsv):
    """Write a df to the specified file"""
    rank_df.to_csv(output_tsv, sep="\t", index=False)

    return


## WRAPPERS
def select_best_per_taxid(main_df):
    """Build a dictionary of taxids with their best genome representative"""

    grouped = main_df.groupby("taxid")
    best_per_taxid = {}

    for taxid in grouped.groups.keys():
        group = grouped.get_group(taxid)
        best_per_taxid[taxid] = get_best_genome_id(group)

    return best_per_taxid


def select_best_for_ranks(
    ncbi_tree, best_per_taxid_df, lineages_df, output_dir
):

    for rank in OFFICIAL_RANKS:
        rank_df = select_best_for_rank(
            rank, ncbi_tree, best_per_taxid_df, lineages_df
        )

        output_tsv = output_dir / pathlib.Path("best_per_{}.tsv".format(rank))
        write_rank_df_to_file(rank_df, output_tsv)
        _logger.info("{} info written in {}".format(rank, output_tsv))

    return


def select_best(
    genome_summary,
    genome_lineage,
    genome_ids,
    ncbi_db,
    thresh,
    output_path,
):

    lineages_df = pd.read_csv(genome_lineage, sep="\t", dtype=column_dtypes)

    # genome_name: is there and gets suffixed with _x, _y
    # taxon_id: is present as taxid in the original df
    # Drop the columns from the lineages_df before merging
    lineages_df = lineages_df.drop(columns=["taxon_id", "genome_name"])

    filtered_data = parse_genome_summary(genome_summary, thresh, genome_ids)

    best_per_taxid = select_best_per_taxid(filtered_data)

    best_df = filtered_data.loc[
        filtered_data.genome_id.isin(best_per_taxid.values())
    ]

    # Create a Series of filepaths
    # local_fp = best_df.apply(lambda x: genome_ids.get(x.genome_id), axis=1)

    # Make a copy to avoid the copy on slice warning from pandas
    final_df = best_df.copy()

    # Append lineages info
    best_per_taxid_df = pd.merge(final_df, lineages_df, on="genome_id")

    # Append the local_fp series to the df
    # final_df["local_fp"] = local_fp

    if not output_path.exists():
        output_path.mkdir()

    # Write the df to the specified output.
    best_per_taxid_tsv = output_path / pathlib.Path("best_per_taxid.tsv")
    best_per_taxid_df.to_csv(best_per_taxid_tsv, sep="\t", index=False)
    _logger.info(
        "Best genomes per taxid written in {}".format(best_per_taxid_tsv)
    )

    _logger.info("Loading NCBI Taxonomy information")
    ncbi = NCBITaxa(dbfile=ncbi_db)

    # Get an array of the taxids covered here
    taxids_used = final_df.taxid.values

    # Create a taxonomy tree for the valid taxids
    ncbi_tree = ncbi.get_topology(taxids_used, intermediate_nodes=True)

    select_best_for_ranks(ncbi_tree, final_df, lineages_df, output_path)
    _logger.info("Done. Results are found in {}".format(output_path))



from concurrent.futures import ProcessPoolExecutor as tPool
import multiprocessing as mp
from Bio import SeqIO
import gzip
import pandas as pd
import pathlib
import logging
import shutil
import sys

_logger = logging.getLogger(__name__)

supported_suffixes = {
    "patric_proteins": "PATRIC.faa",
    "patric_rnas": "PATRIC.frn",
}

SSU_DESCRIPTIONS = ["16S ribosomal RNA", "SSU rRNA"]


def is_gz(fp):
    return fp.name.endswith("gz")


def optionally_compressed_handle(fp, mode):
    if mode == "r" or mode == "rb":
        mode = "rt"
    if mode == "a" or mode == "ab":
        mode = "at"
    if mode == "w" or mode == "wb":
        mode = "wt"

    if is_gz(fp):
        return gzip.open(fp, mode=mode)
    else:
        return open(fp, mode)


def load_index(index_path):
    df = pd.read_csv(index_path, sep="\t", dtype={"genome_id": "string"})
    return df


def select_genome_ids(index_path, on_col):
    """Get a list of genome ids that have a file"""
    df = pd.read_csv(index_path, sep="\t", dtype={"genome_id": "string"})
    genome_ids = df.loc[df[on_col] == 1]["genome_id"].tolist()
    assert len(genome_ids) != 0, _logger.critical("No genome ids found")
    return genome_ids


def generate_file_list(genomes_dir, genome_ids, target, suffixes_dict):
    """Get a list of files that are going to be parsed"""
    if target in suffixes_dict:
        suffix = suffixes_dict[target]
        files_list = [
            genomes_dir / pathlib.Path(f"{genome_id}/{genome_id}.{suffix}")
            for genome_id in genome_ids
        ]
    else:
        _logger.critical("Unsupported filetype for set {}".format(target))
        sys.exit(1)

    return files_list


def is_SSU(record):
    """Return true if the record is a 16S sequence based on description"""
    return any(desc in record.description for desc in SSU_DESCRIPTIONS)


def has_valid_seq(record):
    """Basic check for empties and `1`s"""
    return len(record.seq) > 1


def fasta_reader(fa, q, *filters):
    # TO DO
    # Check packing/unpacking
    # Should this be just a list, even if it's empty?
    """
    Reader worker for the fa file in the specified q(ueue)

    Applies a filter on the sequence length > 1. This
    is there to parse out
    (a) empty sequence strings
    (b) sequences that are only represented as 1

    Parsing is done with Bio.SeqIO.parse()

    The optional `filters` can be any number of callables that can
    be applied to a SeqRecord object. Each should return a single
    boolean True or False, if the record is to be kept. True to
    keep, False to discard. If a record must be kept, all filters
    should return True. If one fails, the record is skipped.


    Arguments:
      fa: Path obj: The Path representation of the fasta to read
      q: Queue obj: A multiprocessing.Manager.Queue() instance
      *filters: callables: Filtering rules that apply a test to
      the record object. They should return a single True or
      False value.


    Return:
        seq_dict: dict: Dictionary that holds seq descriptions as seq ids
        and sequences.
        A 'skipped' key is there to also gather sequence ids that were
        skipped due to the filtering. Of the form
        {
            seq.description : sequence,
                ...,
            'skipped': [seq.description_1, ...]
                ...
        }
    """
    seq_dict = {"skipped": []}
    for record in SeqIO.parse(fa, "fasta"):
        keep = has_valid_seq(record)
        if filters and keep is True:
            keep = all(f(record) for f in filters[0])
        if keep is True:
            seq_dict[record.description] = record.seq
        else:
            seq_dict["skipped"].append(record.id)
            pass

    q.put(seq_dict)

    return seq_dict


def writer(q, out_fp):
    """
    Writer worker that handles writing to files fed from the reader
    in the q(ueue)

    out_fp is gzipped to save space.

    Arguments:
      q: mp.Queue() obj: A queue
      out_fp: Path obj: Path to output.fa.gz

    Return:
      -
    """
    with gzip.open(out_fp, "wt") as fout:
        while 1:
            m = q.get()
            if m == "kill":
                break
            for k, v in m.items():
                if k == "skipped":
                    # Used to handle that
                    pass
                else:
                    fout.write(">{}\n{}\n".format(k, v))
                    fout.flush()


def chunkify(files_list, chunksize=1000):
    """
    Create a list of chunks.

    Each chunk is a list itself, with size `chunksize`.

    Arguments:
      files_list: list: A list of Path objects
      chunksize: int: Size of each chunk

    Return:
      chunks: list: A list of lists. Each nested list has size
        `chunksize`.
    """
    chunks = []
    for i in range(0, len(files_list), chunksize):
        chunk = files_list[i : i + chunksize]
        chunks.append(chunk)
    return chunks


def create_jobs_list(chunks, outdir, *filters):
    # TO DO
    # Figure out the packing/unpacking
    """
    Create a list of dictionaries that hold information for the given
    chunks

    Arguments:
      chunks: list: A list of lists. Each nested list contains the
      filepaths to be processed
      outdir: Path object: The directory where results will be written
      filters: Callables

    Return:
    jobs_list: list: A list of dictionaries that holds information for
    the execution of each chunk. Of the form
      [
        {'chunk_id'  : int,         (0,1,2,...)
         'out_fp'    : Path object, (outdir/chunk_<chunk_id>.fa.gz)
         'fastas'    : list of Path objects,
                       ([PosixPath('path/to/PATRIC.faa'),...])
         'filters'   : list of functions
        }
      ]

    """
    jobs_list = []
    for i, chunk in enumerate(chunks):
        chunk_id = f"chunk_{i}"
        chunk_out = f"{chunk_id}.fa.gz"
        out_fp = outdir / pathlib.Path(chunk_out)
        # chunk_skipped = f"{chunk_id}.skipped.txt"
        chunk_fastas = chunk
        chunk_dict = {
            "chunk_id": chunk_id,
            "fastas": chunk_fastas,
            "out_fp": out_fp,
            # Should there be an if filters or if len(filters) != 0 ?
            "filters": [f for f in filters],
        }

        jobs_list.append(chunk_dict)
    return jobs_list


# https://stackoverflow.com/a/13530258
def process_chunk(chunk_dict):
    """
    Multiprocessing of a single chunk.

    This spawns 4 processes: 1 is dedicated to writing and 3 are reading.
    This is to ensure that the file being written is locked.
    """
    manager = mp.Manager()
    q = manager.Queue()

    # TO DO
    # Avoid hardcoding this value
    with mp.Pool(4) as mpool:
        # 1 dedicated process to write
        watcher = mpool.apply_async(writer, (q, chunk_dict["out_fp"]))

        # The rest 3 are reading
        jobs = []
        for fa in chunk_dict["fastas"]:
            job = mpool.apply_async(
                fasta_reader, (fa, q, chunk_dict["filters"])
            )
            jobs.append(job)

        # Collect results
        for job in jobs:
            job.get()

        q.put("kill")
        mpool.close()
        mpool.join()


def collect_sequences(files_list, outdir, *filters, nthreads=2):
    """Start nthreads that each spawns a multiprocessing pool of 4

    Yeah, this is weird. Probably a threads-only based thing would
    be better
    """
    chunks = chunkify(files_list)

    job_dicts = create_jobs_list(chunks, outdir, *filters)

    # Restrict cpu usage
    max_cpus = mp.cpu_count()
    if nthreads * 4 > max_cpus:
        _logger.info("Too many threads provided ( {} )".format(nthreads))
        _logger.info("This is an implementation detail that should be fixed")
        use_threads = nthreads // 4
        _logger.info("Resorting to {}".format(use_threads))
    else:
        use_threads = nthreads
        _logger.info("Using {} threads".format(use_threads))

    # tPool is concurrent.futures.PoolExecutor
    # It allows for children processes
    # to spawn their own pools, even multiprocessing.Pool
    # see https://stackoverflow.com/a/61470465
    with tPool(use_threads) as tpool:
        counter = 0
        for i in range(0, len(job_dicts), nthreads):
            current_jobs = job_dicts[i : i + nthreads]
            a = tpool.map(process_chunk, current_jobs)
            for _ in a:
                counter += 1
            _logger.info(
                "Finished {} / {} chunks".format(counter, len(chunks))
            )
    return


def concatenate_chunk_files(chunks_dir, output_path, cleanup=True):
    """Gather all files in one"""
    _logger.info("Concatenatating all sequence files")
    files_counter = 0
    seq_counter = 0
    with optionally_compressed_handle(output_path, "w") as fout:
        for f in chunks_dir.iterdir():
            files_counter += 1
            with optionally_compressed_handle(f, "r") as fin:
                for record in SeqIO.parse(fin, "fasta"):
                    SeqIO.write(record, fout, "fasta")
                    seq_counter += 1
    if cleanup is True:
        _logger.info("Removing {}".format(chunks_dir.resolve()))
        shutil.rmtree(chunks_dir)

    return files_counter, seq_counter



import os
import logging
import pandas as pd
import hashlib
import pathlib
from io import BytesIO


# CONSTANTS

PATRIC_FTP = "ftp.patricbrc.org"

RELEASE_NOTES_FILES = {
    "genome_summary",
    "genome_metadata",
    "genome_lineage",
    "PATRIC_genomes_AMR.txt",
}

_logger = logging.getLogger(__name__)

# HELPERS

def get_remote_dir_timestamps(ftp_handle, remote_dir_name, skip_dirs=True):
    """
    Parse the remote timestamps.

    Create a dict that holds file names as keys and timestamps as values
    Timestamp is formatted as 'YYYYMMDDHHMMSS'

    Positional arguments:
      - ftp_handle: ftplib.FTP: A working ftp connection
      - remote_dir_name: string: A relative path to the ftp root.
        e.g. 'RELEASE_NOTES/genome_summary'

    Keyword arguments:
      - skip_dirs: bool: If this is True directories, are skipped.

    Return:
      - timestamps_dict: dict: {fname: timestamp, ... }. `fname` as read from
      the ftp. `timestamp` is formatted as 'YYYYMMDDHHMMSS'
        e.g. { '1234.5.fna' : '20200112235902', ... }
    """
    timestamps_dict = {}
    for entry in ftp_handle.mlsd(remote_dir_name, facts=["modify", "type"]):
        if (entry[1]["type"] != "file") and (skip_dirs is True):
            pass
        else:
            timestamps_dict[entry[0]] = {"ftp_mdtm": entry[1]["modify"]}

    return timestamps_dict


def get_remote_file_md5(ftp_handle, remote_file_name):
    """Store the md5sum of a file and its contents in memory"""
    _logger.debug(
        "Calculating md5sum for file: {} from remote".format(remote_file_name)
    )
    content = BytesIO()
    ftp_handle.retrbinary(f"RETR {remote_file_name}", content.write)
    content_string = content.getvalue()
    md5 = hashlib.md5(content_string).hexdigest()
    return md5, content_string


def get_local_info(a_dir):
    """Get records from the ftp_info.tsv if it exists"""
    dir_info = None
    info_tsv = a_dir / pathlib.Path("ftp_info.tsv")
    if info_tsv.exists():
        dir_info = pd.read_csv(
            info_tsv, sep="\t", dtype={"ftp_mdtm": "string"}, index_col="fname"
        )
        dir_info = dir_info.to_dict(orient="index")
    else:
        _logger.debug(
            "No mod time information found for {}".format(a_dir.resolve())
        )
    return dir_info


def get_missing_files(a_dir, files_list):
    """Given a list of filenames, scan the dir for their existence"""
    missing_files = []
    for f in files_list:
        fp = a_dir / pathlib.Path(f)
        if not fp.exists():
            missing_files.append(f)

    if len(missing_files) != 0:
        _logger.debug(
            "Directory: {} is missing file(s): {}".format(
                a_dir.resolve(), ",".join(missing_files)
            )
        )
    else:
        logging.debug(
            "Directory: {} contains all files".format(a_dir.resolve())
        )

    return set(missing_files)


def md5(fp):
    """Calculate md5sum for a file"""
    content = fp.read_text().encode()
    return hashlib.md5(content).hexdigest()


def get_dir_md5(a_dir, files_list):
    md5_data = {}
    for f in a_dir.iterdir():
        if f.name in files_list:
            md5_data[f.name] = {"md5": md5(f)}
    return md5_data


def dir_is_empty(dirpath):
    with os.scandir(dirpath) as it:
        return not any(it)


def download_single_file(ftp_handle, remote_fn, local_fp):

    """
    Download one file from the ftp.

    Given a valid ftplib.FTP connection, download the remote_fn
    and store it in the local_fp.

    Positional arguments:
      - ftp_handle: ftplib.FTP:
      - remote_fn: str: Absolute paths or relative to the root dir.
          e.g. "'/genomes/1234.5' or 'genomes/1234.5'
            but not '1234.5'
      -  local_fp: pathlib.Path: Path where the remote is stored

    Return:
      - None

    """
    _logger.debug("Downloading {} in {}".format(remote_fn, local_fp))
    with open(local_fp, "wb") as fout:
        ftp_handle.retrbinary("RETR {}".format(remote_fn), fout.write)


def load_cached_genomes(processed_txt):
    with open(processed_txt, "r") as fin:
        processed_genomes = [line.strip() for line in fin]
    return processed_genomes


def genomes_from_summary(genome_summary):
    genome_ids = pd.read_csv(
        genome_summary,
        sep="\t",
        usecols=["genome_id"],
        dtype={"genome_id": "string"},
    )
    genome_ids = genome_ids["genome_id"].unique()

    _logger.info(
        "Loaded {} genome ids from {}".format(len(genome_ids), genome_summary)
    )

    return genome_ids


def filter_files_on_mdtm(remote_tstamps, local_tstamps=None):
    """
    Compare two timestamps dictionaries for files for files changed

    Filenames are basenames

    Positional arguments:
      - remote_tstamps: dict: A dictionary
          {filename: {'ftp_mdtm' : timestamp, ... }, ...}
      - local_tstamps: dict or None: A dictionary
          {filename: {'ftp_mdtm': timestamp} ... } , ...}
      If it is `None`, all files from `remote_tstamps` are returned

    Return:
      - targets_list: list: A list of filenames. Can be empty if nothing has changed.
    """
    targets_list = []
    if local_tstamps is not None:
        for f in local_tstamps:
            if local_tstamps[f]["ftp_mdtm"] != remote_tstamps[f]["ftp_mdtm"]:
                targets_list.append(f)
    else:
        targets_list = list(remote_tstamps.keys())

    return targets_list


def download_genome_targets(ftp_handle, local_dir, targets_list):
    """Download the list of target filenames to the local dir"""
    genome_id = local_dir.name
    for f in targets_list:
        remote_fname = f"genomes/{genome_id}/{f}"
        local_fpath =  local_dir / pathlib.Path(f)
        download_single_file(ftp_handle, remote_fname, local_fpath)




"""
References:
    - https://setuptools.readthedocs.io/en/latest/userguide/entry_point.html
    - https://pip.pypa.io/en/stable/reference/pip_install
"""

import click
import logging
import pathlib
import sys

from cirtap import __version__
from cirtap.mirror import (
    check_release_dir,
    check_cache_dir,
    create_genome_jobs,
)
from cirtap.mirror import mirror_genomes_dir
from cirtap.index import contents, all_data, write_index
from cirtap.collect import supported_suffixes, is_SSU
from cirtap.collect import (
    select_genome_ids,
    generate_file_list,
    collect_sequences,
    concatenate_chunk_files,
)
from cirtap.best import select_best
from cirtap.pack import pack_genome_data
from cirtap.mailer import send_start_mail, send_exit_mail

__author__ = "papanikos"
__copyright__ = "papanikos"
__license__ = "MIT"

_logger = logging.getLogger(__name__)


def setup_logging(loglevel, logfile):
    """Setup basic logging

    Args:
      loglevel: str: minimum loglevel for emitting messages
    """
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: {}".format(loglevel))

    hs = [logging.StreamHandler(stream=sys.stderr)]
    if logfile:
        filelogger = logging.FileHandler(logfile, mode="w")
        hs.append(filelogger)

    logformat = "[%(asctime)s - %(levelname)s:%(name)s] %(message)s"

    logging.basicConfig(
        level=numeric_level,
        handlers=hs,
        format=logformat,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(__version__)
def cli():
    """Run `cirtap COMMAND -h` for subcommand help"""
    pass


@click.command()
@click.argument("db_dir", required=True, type=pathlib.Path)
@click.option(
    "--cache-dir",
    type=pathlib.Path,
    help="Directory where cirtap will store some info for its execution. "
    "Subsequent executions rely on it so be careful when you delete",
    required=False,
)
@click.option(
    "-j",
    "--jobs",
    default=1,
    show_default=True,
    help="Number of parallel processes to start for downloading",
)
@click.option(
    "--skip-release-check",
    default=False,
    is_flag=True,
    show_default=True,
    help="Skip checking for RELEASE_NOTES based updates",
)
@click.option(
    "--skip-processed-genomes",
    default=False,
    show_default=True,
    is_flag=True,
    help="Skip checks for already processed genomes as found in the cache.",
)
@click.option(
    "--force-check",
    is_flag=True,
    default=False,
    show_default=True,
    help="Force update the genomes directory regardless of RELEASE_NOTES "
    "outcome",
)
@click.option(
    "-r",
    "--resume",
    is_flag=True,
    default=False,
    show_default=True,
    help="Use this to set both --skip-release-check and "
    "--skip-processed-genomes on. "
    "Useful for resuming a failed run",
)
@click.option(
    "--loglevel",
    default="INFO",
    help="Define loglevel",
    show_default=True,
    required=False,
)
@click.option(
    "--archive-notes",
    is_flag=True,
    default=False,
    show_default=True,
    help="Create an tar.gz archive of the RELEASE_NOTES files in the DB_DIR",
)
@click.option(
    "--notify",
    type=click.STRING,
    required=False,
    help="Comma (,) separated list of emails provided as a string. E.g. "
    "'user1@mail.com,user2@anothermail.com'",
)
@click.option(
    "--progress",
    is_flag=True,
    required=False,
    show_default=True,
    help="(Experimental) Print a progress bar when downloading genomes. This "
    "option cannot be set with `--loglevel debug`. If they are both supplied, "
    "progress will not be shown and the more descriptive debugging messages "
    "will be printed to stderr instead",
)
@click.option(
    "--logfile",
    help="Write logging information in this file",
    show_default=True,
    required=False,
)
def mirror(
    db_dir,
    loglevel,
    cache_dir,
    skip_release_check,
    jobs,
    skip_processed_genomes,
    notify,
    archive_notes,
    resume,
    force_check,
    progress,
    logfile,
):
    """Mirror all data from ftp.patricbrc.org in the specified DB_DIR"""

    setup_logging(loglevel, logfile)
    _logger.info("Full command: {}".format(" ".join(sys.argv[:])))
    _logger.info("Version: {}".format(__version__))

    if progress and (loglevel == "debug"):
        _logger.info(
            "Unsetting `progress` option because it conflicts with "
            "`loglevel debug`"
        )
        progress = False

    release_notes_dir = db_dir / pathlib.Path("RELEASE_NOTES")
    genomes_dir = db_dir / pathlib.Path("genomes")

    if not db_dir.exists():
        _logger.info("Fresh mirror in: {}".format(db_dir.resolve()))
        db_dir.mkdir(parents=True)
        release_notes_dir.mkdir()
        genomes_dir.mkdir()

    # Load already processed genomes from cache if found
    # Otherwise processed_genomes is an empty set
    if not cache_dir:
        cache_dir = db_dir / pathlib.Path(".cache")
    processed_genomes = check_cache_dir(cache_dir)

    check_genomes = True
    if not skip_release_check and not resume:
        _logger.info("Checking RELEASE_NOTES status".format(release_notes_dir))
        check_genomes = check_release_dir(
            release_notes_dir, archive=archive_notes
        )

    genome_summary = release_notes_dir / pathlib.Path("genome_summary")

    # Create a list of jobs that can be multiprocessed
    genome_jobs = create_genome_jobs(
        genome_summary, genomes_dir, processed_genomes
    )

    # Do not check already existing genome_ids retrieved from the cache
    # These will not be submitted in the main mirror step
    # Speeds up re-executions if things failed the first time
    # Assumes that an update has not occured in between
    if skip_processed_genomes or resume:
        genome_jobs = [
            job for job in genome_jobs if job not in processed_genomes
        ]

    # Try to notify but don't try too hard
    if notify:
        recipients = [str(i).strip() for i in notify.split(",")]
        try:
            send_start_mail(
                recipients,
                db_dir,
                len(genome_jobs),
            )
        except:
            _logger.debug("Failed to send email")

    # Testing
    # ten_targets = [
    #    "100053.5",
    #    "100.11",
    #    "100053.4",
    #    "100.9",
    #    "1123738.3",
    #    "1000562.3",
    #    "100053.8",
    #    "469009.4",
    #    "1309411.5",
    #    "100053.6",
    # ]
    # genome_jobs = [job for job in genome_jobs if job in ten_targets]

    try:
        if (
            len(genome_jobs) != 0 and check_genomes is True
        ) or force_check is True:
            finished_jobs = mirror_genomes_dir(
                genome_jobs, genomes_dir, jobs, progress_bar=progress
            )
            new_genomes_processed = processed_genomes.union(set(finished_jobs))
            processed_genomes_txt = cache_dir / pathlib.Path(
                "processed_genomes.txt"
            )
            # Re-write the file with all new and old ids
            with open(processed_genomes_txt, "w") as fout:
                for genome_id in new_genomes_processed:
                    fout.write(f"{genome_id}\n")
        else:
            _logger.info(
                "All genomes for this version of RELEASE_NOTES seem "
                "to have been properly processed"
            )
    except Exception as e:
        if notify:
            recipients = [str(i).strip() for i in notify.split(",")]
            send_exit_mail(recipients, str(e))

        _logger.critical("Mirror job failed with \n{}".format(e))
        raise

    if notify:
        recipients = [str(i).strip() for i in notify.split(",")]
        send_exit_mail(recipients)


@click.command()
@click.argument(
    "genomes-dir",
    required=True,
    type=pathlib.Path,
)
@click.argument(
    "output-index",
    required=True,
    type=pathlib.Path,
)
@click.option(
    "--loglevel",
    default="INFO",
    help="Define loglevel",
    show_default=True,
    required=False,
)
@click.option(
    "--logfile",
    help="Write logging information in this file",
    show_default=True,
    required=False,
)
@click.option(
    "-j",
    "--jobs",
    default=1,
    show_default=True,
    help="Number of parallel reads to execute. Speeds things up when "
    "iterating over all the data dirs",
)
def index(genomes_dir, output_index, loglevel, logfile, jobs):
    """Create an index of contents for all directories

    This can be useful for generating valid paths before gathering inputs.
    The output_index is a tab-separated file with each column representing
    the files that are expected to be found for a genome that has full
    information available from both PATRIC and RefSeq. If any of the files
    is missing the value is 0.

    genomes_dir: The location where all data is stored

    output_index:  The files to write all the info in

    """
    setup_logging(loglevel, logfile)
    all_records = all_data(genomes_dir, contents, jobs)
    write_index(all_records, output_index)
    return


@click.command()
@click.argument("genomes-dir", required=True, type=pathlib.Path)
@click.argument("output-path", required=True, type=pathlib.Path)
@click.option("-i", "--index-path", required=True, type=pathlib.Path)
@click.option(
    "-t",
    "--target-set",
    type=click.Choice(["SSU", "proteins"]),
    help="Sequence set to create. One of `SSU` (based on .PATRIC.frn) and "
    "`proteins` (based on .PATRIC.faa)",
)
@click.option(
    "-j",
    "--jobs",
    help="Parallel jobs to run",
    default=1,
    show_default=True,
    type=int,
)
@click.option(
    "--cleanup",
    help="Remove all intermediate files produced",
    is_flag=True,
    default=True,
)
@click.option(
    "--loglevel",
    default="INFO",
    help="Define loglevel",
    show_default=True,
    required=False,
)
@click.option(
    "--logfile",
    help="Write logging information in this file",
    show_default=True,
    required=False,
)
def collect(
    genomes_dir,
    index_path,
    target_set,
    output_path,
    jobs,
    loglevel,
    logfile,
    cleanup,
):
    """
    Create sequence sets based on the installed files

    Only 16S rRNA seqs and proteins are supported for now.

    **ATTENTION**
    Due to the way things are implemented any jobs number you
    supply will be multiplied by 4.
    """
    setup_logging(loglevel, logfile)

    filters = []
    if target_set == "SSU":
        target = "patric_rnas"
        filters.append(is_SSU)
    elif target_set == "proteins":
        target = "patric_proteins"
    else:
        _logger.critical("Case of {} not covered".format(target_set))
        sys.exit(1)

    _logger.info("Reading index information")
    genome_ids = select_genome_ids(index_path, target)

    _logger.info("Generating files list")
    files_list = generate_file_list(
        genomes_dir, genome_ids, target, supported_suffixes
    )
    _logger.info(
        "Sequences will be collected from {} files".format(len(files_list))
    )

    tmp_dir = output_path.parent / pathlib.Path("tmp")
    tmp_dir.mkdir(exist_ok=True)

    collect_sequences(files_list, tmp_dir, *filters, nthreads=jobs)

    files_no, seqs_no = concatenate_chunk_files(tmp_dir, output_path, cleanup)
    _logger.info(
        "Collected {} sequences from {} files".format(seqs_no, files_no)
    )


@click.command()
@click.argument(
    "output-path",
    type=pathlib.Path,
)
@click.option(
    "-i",
    "--index-path",
    type=pathlib.Path,
    required=True,
    help="Path to the index file",
)
@click.option(
    "-d",
    "--db-dir",
    type=pathlib.Path,
    required=True,
    help="Path to the local mirror. Must contain a `RELEASE_NOTES` directory "
    "and a `genomes` directory",
)
@click.option(
    "--thresh",
    type=int,
    help="Integer threshold for including a genome based on "
    "completeness and contamination stats. This is useed as "
    "completeness - 5*contamination > thresh",
    default=70,
    show_default=True,
)
@click.option(
    "--ncbi-db",
    type=pathlib.Path,
    help="Path to the taxa.sqlite created by ete3",
    default=pathlib.Path.home() / pathlib.Path(".etetoolkit/taxa.sqlite"),
    show_default=True,
)
@click.option(
    "--loglevel",
    default="INFO",
    help="Define loglevel",
    show_default=True,
    required=False,
)
@click.option(
    "--logfile",
    help="Write logging information in this file",
    show_default=True,
    required=False,
)
def best(db_dir, index_path, output_path, thresh, ncbi_db, loglevel, logfile):
    """
    Select best genomes based on stats retrieved from genome_summary

    Results are written in the provided output_path and include tables
    containing genome ids with their stats and lineages attached
    """
    setup_logging(loglevel, logfile)

    genomes_dir = db_dir / pathlib.Path("genomes")
    notes_dir = db_dir / pathlib.Path("RELEASE_NOTES")
    genome_summary = notes_dir / pathlib.Path("genome_summary")
    genome_lineage = notes_dir / pathlib.Path("genome_lineage")
    if not all(
        p.resolve().exists()
        for p in [notes_dir, genome_summary, genome_lineage, genomes_dir]
    ):
        _logger.error("Please provide a valid db path. ")
        _logger.error(
            "This must contain "
            "a RELEASE_NOTES directory with `genome_summary` and "
            "`genome_lineage` files and a separate `genomes` dir "
            "where all genome files are stored"
        )
        sys.exit(1)

    genome_ids = select_genome_ids(index_path, "patric_genome")
    _logger.info(
        "Loaded {} valid genome ids from the index".format(len(genome_ids))
    )

    select_best(
        genome_summary,
        genome_lineage,
        genome_ids,
        ncbi_db,
        thresh,
        output_path,
    )
    _logger.info("Done! Results are in: {}".format(output_path.resolve()))


@click.command()
@click.option(
    "-g",
    "--genomes-dir",
    type=pathlib.Path,
    required=True,
    help="Path to the genomes directory containing all data",
)
@click.option(
    "-i",
    "--input-list",
    type=pathlib.Path,
    help="Single column file containing one genome id per line",
    required=True,
)
@click.option(
    "-o",
    "--output",
    type=pathlib.Path,
    help="Output gzipped file to write. For now only gzip compression is "
    "supported",
    required=True,
)
@click.option(
    "--loglevel",
    default="INFO",
    help="Define loglevel",
    show_default=True,
    required=False,
)
@click.option(
    "--logfile",
    help="Write logging information in this file",
    show_default=True,
    required=False,
)
def pack(genomes_dir, input_list, output, loglevel, logfile):
    """Create a gzipped tar archive from a list of genome ids in a file

    For now the output specified is a tar.gz file, even if you don't name it
    as such.
    """
    setup_logging(loglevel, logfile)
    pack_genome_data(genomes_dir, input_list, output)


cli.add_command(mirror)
cli.add_command(index)
cli.add_command(collect)
cli.add_command(best)
cli.add_command(pack)


if __name__ == "__main__":
    cli()



import tarfile
import pathlib
import logging

_logger = logging.getLogger(__name__)


def genome_ids_from_text(txt_file):
    """Parse single column text file to a set"""
    genome_ids = []
    with open(txt_file, "r") as fin:
        for line in fin:
            genome_id = line.strip()
            genome_ids.append(genome_id)
    genome_ids_set = set(genome_ids)
    _logger.info(
        "Read {} unique ids from {} lines".format(
            len(genome_ids_set), len(genome_ids)
        )
    )

    return genome_ids_set


def pack_genome_data(genomes_dir, ids_text, tar_out):
    """Create a gzipped tarfile from the ids in the text file"""

    genome_ids = genome_ids_from_text(ids_text)

    with tarfile.open(tar_out, "w:gz") as fout:
        for genome_id in genome_ids:
            full_path = genomes_dir / pathlib.Path(genome_id)
            _logger.debug(full_path)
            if full_path.exists():
                for f in full_path.iterdir():
                    leaf = f.name
                    tar_name = f"genomes/{genome_id}/{leaf}"
                    _logger.debug("Compressing {}".format(f.resolve()))
                    fout.add(f, arcname=tar_name)
            else:
                _logger.warning(f"No data found for genome id {genome_id}")

    return



from functools import partial
import ftplib
import logging
import multiprocessing as mp
import pandas as pd
import pathlib
import shutil
import time
from tqdm import tqdm

from .common import PATRIC_FTP, RELEASE_NOTES_FILES
from .common import get_missing_files, get_dir_md5, get_remote_dir_timestamps
from .common import load_cached_genomes, genomes_from_summary
from .common import get_local_info, get_remote_file_md5
from .common import filter_files_on_mdtm, download_genome_targets

_logger = logging.getLogger(__name__)


def set_remote_version_on_summary():
    """Use the year and month in YYYYMM format as a version"""
    with ftplib.FTP(PATRIC_FTP) as ftp:
        ftp.login()
        release_notes_times = get_remote_dir_timestamps(ftp, "RELEASE_NOTES")
    remote_version = release_notes_times["genome_summary"]["ftp_mdtm"][:6]

    _logger.debug("Remote version is set to {}".format(remote_version))

    return remote_version


def set_local_version(release_dir):
    """Try to infer version in YYYYMM format from existing data"""
    local_version = None
    local_timestamps = get_local_info(release_dir)
    if local_timestamps:
        try:
            local_version = local_timestamps["genome_summary"]["ftp_mdtm"][:6]
        except KeyError:
            pass

    local_version_file = release_dir / pathlib.Path("VERSION")
    if local_version_file.exists():
        local_version = local_version_file.read_text()[:6]

    _logger.debug("Local version is set to {}".format(local_version))
    return local_version


def archive_notes(release_dir, version_suffix=None):
    """Create an archive of all RELEASE_NOTES contents in the parent dir"""

    db_dir = release_dir.parent.resolve()
    if version_suffix:
        basename = "{}.RELEASE_NOTES.bkp".format(version_suffix)
    else:
        basename = "REELASE_NOTES.bkp"
    archive_name = db_dir / pathlib.Path(basename)
    root_dir = db_dir.resolve()
    base_dir = "RELEASE_NOTES"

    shutil.make_archive(
        archive_name, "gztar", root_dir=root_dir, base_dir=base_dir
    )
    return


def check_release_dir(release_dir, archive=True):
    """Check for updates and download if necessary"""

    remote_version = set_remote_version_on_summary()
    local_version = set_local_version(release_dir)

    if remote_version == local_version:
        _logger.info("Versions based on genome_summary are the same")
    if not local_version:
        local_version = "unknown_version"

    missing_files = get_missing_files(release_dir, RELEASE_NOTES_FILES)

    present_files = RELEASE_NOTES_FILES.difference(missing_files)

    if (len(present_files) != 0) and (archive is True):
        _logger.info("Archiving existing RELEASE_NOTES")
        archive_notes(release_dir, version_suffix=local_version)

    # Set the new version
    version_txt = release_dir / pathlib.Path("VERSION")
    version_txt.write_text(remote_version)

    local_md5s = get_dir_md5(release_dir, RELEASE_NOTES_FILES)

    check_genomes = False
    with ftplib.FTP(PATRIC_FTP) as ftp:
        ftp.login()

        for f in RELEASE_NOTES_FILES:
            remote_name = f"RELEASE_NOTES/{f}"
            remote_md5, contents = get_remote_file_md5(ftp, remote_name)
            local_fp = release_dir / pathlib.Path(f)
            if f in present_files:
                if remote_md5 != local_md5s[f]["md5"]:
                    _logger.info("Updating file: {}".format(f))
                    local_fp.write_bytes(contents)
                    check_genomes = True
                else:
                    _logger.info("File: {} is up to date".format(f))

            elif f in missing_files:
                _logger.info("Fetching file: {}".format(f))
                local_fp.write_bytes(contents)
                check_genomes = True

            else:
                _logger.critical("WHAT HAPPENED with {}?".format(f))
                raise

    return check_genomes


def check_cache_dir(cache_dir):
    processed_genomes = []
    if not cache_dir.exists():
        _logger.debug("No cache found. Creating it at {}".format(cache_dir))
        cache_dir.mkdir()
        processed_txt = cache_dir / pathlib.Path("processed_genomes.txt")
    else:
        try:
            processed_txt = cache_dir / pathlib.Path("processed_genomes.txt")
            processed_genomes = load_cached_genomes(processed_txt)
        except FileNotFoundError:
            _logger.debug(
                "No processed_genomes.txt found in {}".format(
                    cache_dir.resolve()
                )
            )
    return set(processed_genomes)


# @tenacity.retry(
#    retry=tenacity.retry.retry_if_exception_type(ftplib.all_errors),
#    before=tenacity.before.before_log(_logger, log_level=logging.DEBUG),
#    wait=tenacity.wait.wait_exponential(multiplier=1, min=10, max=60),
#    stop=tenacity.stop.stop_after_attempt(3),
# )
def sync_single_dir(genomes_dir, genome_id, attempts=3, write_info=True):
    """Download/update all info for the genome id"""

    remote_dirname = f"genomes/{genome_id}"

    local_dirpath = genomes_dir / pathlib.Path(genome_id)
    # get_local_info should returns None if either the local_dirpath is
    # not there or the file itself
    local_info = get_local_info(local_dirpath)

    # This loop should raise after attempt 3 and break the thing
    # no matter what
    # So no genome id will be there
    for attempt in range(1, attempts + 1):
        try:
            with ftplib.FTP(PATRIC_FTP) as ftp:
                ftp.login()
                remote_info = get_remote_dir_timestamps(ftp, remote_dirname)
                targets = filter_files_on_mdtm(remote_info, local_info)
                missing_files = get_missing_files(
                    local_dirpath, list(remote_info.keys())
                )

                if len(missing_files) != 0:
                    targets.extend(missing_files)

                if len(targets) != 0:
                    # Create the dir only if there is something to download
                    local_dirpath.mkdir(exist_ok=True)
                    download_genome_targets(ftp, local_dirpath, targets)
                else:
                    _logger.debug("{} is up to date".format(genome_id))

            if write_info is True:
                ftp_info_tsv = local_dirpath / pathlib.Path("ftp_info.tsv")
                modtimes_df = pd.DataFrame.from_dict(
                    remote_info, orient="index"
                )
                modtimes_df.index.name = "fname"
                modtimes_df.to_csv(ftp_info_tsv, sep="\t")

            break

        # Specific handling of the '550 - No such file or directory'
        # Breaks and returns the genome_id
        except ftplib.error_perm as ftp_err:
            if str(ftp_err).startswith("550"):
                _logger.debug(
                    "Skipping {} - No remote dir found".format(genome_id)
                )
                break

        # Try to catch CTRL-C if user doesn't want to proceed
        except KeyboardInterrupt:
            _logger.debug("Ctrl+C signal detected")
            _logger.debug(
                "Removing directory that might contain corrupted files "
                "at {}".format(local_dirpath)
            )
            shutil.rmtree(local_dirpath.resolve())
            break

        # For any other exception e
        # Retry until number of specified attempts and raise eventually
        # Capture text in the generic Exception class
        except Exception as e:
            if attempt == attempts:
                _logger.debug("Failed syncing {}".format(genome_id))
                _logger.debug(
                    "Removing directory that might contain corrupted files "
                    "at {}".format(local_dirpath)
                )
                shutil.rmtree(local_dirpath.resolve())
                raise
            else:
                _logger.debug("Failed syncing  {}".format(genome_id))
                _logger.debug("Error was : {}".format(e))
                attempt += 1
                _logger.debug(
                    "Sleeping for {} s before retrying".format(attempt * 60)
                )
                time.sleep(attempt * 60)

    return genome_id


def create_genome_jobs(genome_summary, genomes_dir, processed_genomes=None):

    all_genomes = genomes_from_summary(genome_summary)
    all_genome_jobs = []

    if processed_genomes:
        _logger.debug(
            "{} genomes will be skipped".format(len(processed_genomes))
        )
        all_genomes = sorted(
            [i for i in all_genomes if i not in processed_genomes]
        )
        _logger.debug("Continuing with {} genomes".format(len(all_genomes)))

    all_genome_jobs = all_genomes
    #    all_genome_jobs = [
    #        (genome_id, pathlib.Path(genomes_dir) / pathlib.Path(genome_id))
    #        for genome_id in all_genomes
    #    ]
    #
    return all_genome_jobs


def mirror_genomes_dir(
    all_genome_jobs, local_genomes_dir, procs=1, progress_bar=True
):

    parallel_sync = partial(
        sync_single_dir, local_genomes_dir, write_info=True, retries=3
    )

    results = []
    with mp.Pool(processes=procs) as pool:
        if progress_bar:
            pbar = tqdm(total=len(all_genome_jobs))
            for res in pool.imap_unordered(parallel_sync, all_genome_jobs):
                pbar.update()
                results.append(res)
        else:
            for res in pool.imap_unordered(parallel_sync, all_genome_jobs):
                results.append(res)
    return results



#!/usr/bin/env python

import logging
import pathlib
import pandas as pd
import multiprocessing as mp

_logger = logging.getLogger(__name__)

## A "complete" dir listing
# 511145.12.fna
# 511145.12.PATRIC.faa
# 511145.12.PATRIC.features.tab
# 511145.12.PATRIC.ffn
# 511145.12.PATRIC.frn
# 511145.12.PATRIC.gff
# 511145.12.PATRIC.pathway.tab
# 511145.12.PATRIC.spgene.tab
# 511145.12.PATRIC.subsystem.tab
# 511145.12.RefSeq.cds.tab
# 511145.12.RefSeq.faa
# 511145.12.RefSeq.ffn
# 511145.12.RefSeq.frn
# 511145.12.RefSeq.gbf
# 511145.12.RefSeq.gff
# 511145.12.RefSeq.pathway.tab
# 511145.12.RefSeq.rna.tab

# TO DO
# This is a bit awkward to use
contents = {
    "genome": {"fna": "patric_genome"},
    "PATRIC": {
        "faa": "patric_proteins",
        "features.tab": "patric_features",
        "ffn": "patric_functions",
        "frn": "patric_rnas",
        "gff": "patric_annotations",
        "pathway.tab": "patric_patwhays",
        "spgene.tab": "patric_spgenes",
        "subsystem.tab": "patric_subsystems",
    },
    "RefSeq": {
        "cds.tab": "refseq_cds",
        "faa": "refseq_proteins",
        "features.tab": "refseq_features",
        "ffn": "refseq_functions",
        "frn": "refseq_rnas",
        "gbf": "refseq_genbank",
        "gff": "refseq_annotations",
        "pathway.tab": "refseq_patwhays",
        "rna.tab": "refseq_rna_tab",
    },
}


# def construct_filename(genome_id, db, f):
#    db_suffix = contents["db"][db]["name"]
#    file_suffix = contents["db"][db]["files"][f]
#    return f"{genome_id}.{db_suffix}.{file_suffix}"


def genome_data(genome_dir, contents):
    """Create a dict of presence (1) | absence (0) of files in a dir"""
    genome_id = genome_dir.name
    genome_data = {"genome_id": genome_id}

    # Required for cases like 7227.4 where an ht2.tar was there
    valid_suffixes = []
    for d in contents.keys():
        for dk in contents[d].keys():
            if dk not in valid_suffixes:
                valid_suffixes.append(dk)

    _logger.debug("Processing dir: {}".format(genome_dir))
    for f in genome_dir.iterdir():
        fname = f.name
        fname_fields = fname.split(".")

        if fname == "ftp_info.tsv":
            continue

        if fname_fields[-1] == "fna":
            col_name = contents["genome"]["fna"]
            genome_data["patric_genome"] = 1
        else:
            db = fname_fields[2]
            suffix = ".".join(fname_fields[3:])
            if suffix in valid_suffixes:
                if db == "RefSeq":
                    col_name = contents["RefSeq"][suffix]
                else:
                    col_name = contents["PATRIC"][suffix]

                genome_data[col_name] = 1

    # TO DO
    # Change the contents dic to something more sane
    ## For the files that were not found
    for d in contents.keys():
        for v in contents[d].values():
            if v not in genome_data:
                genome_data[v] = 0

    return genome_data


def all_data(genomes_dir, contents_dic, jobs=1):
    """Parallelizable read of all directories in a list of dicts"""
    all_records = []
    _logger.info("Loading directories")
    dir_jobs = [(g, contents_dic) for g in genomes_dir.iterdir()]
    _logger.info("Reading information")
    with mp.Pool(processes=jobs) as pool:
        all_records = pool.starmap(genome_data, dir_jobs)

    return all_records


def write_index(records_list, output_index):
    """Create and write a pandas df to the specified file"""
    # Grab the first record dict
    some_record = records_list[0]
    # Map all values to integers
    dtypes = {k: "int64" for k in some_record.keys()}
    # Except for genome id that needs to be a string
    dtypes["genome_id"] = "string"
    index_df = pd.DataFrame.from_records(records_list)
    index_df = index_df.astype(dtypes)

    # Show missing genomes first
    index_df = index_df.sort_values(by="patric_genome")

    # Reorder the columns so that genome id and genome show first
    cols_reordered = ["genome_id", "patric_genome"]
    for col in index_df.columns.tolist():
        if col not in cols_reordered:
            cols_reordered.append(col)
    index_df = index_df[cols_reordered]

    # Write it
    index_df.to_csv(output_index, sep="\t", index=False)
